#
#  ktool | ktool
#  dyld.py
#
#  This file includes a lot of utilities, classes, and abstractions
#  designed for replicating certain functionality within dyld.
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2021.
#
import math
from collections import namedtuple
from enum import Enum

from kmacho import (
    MH_FLAGS,
    MH_FILETYPE,
    LOAD_COMMAND,
    BINDING_OPCODE,
    LOAD_COMMAND_MAP
)
from kmacho.structs import *
from kmacho.base import Constructable
from .macho import _VirtualMemoryMap, Segment
from .util import log, macho_is_malformed
from .exceptions import *


class Dyld:
    """
    This is a static class containing several methods for, essentially, recreating the functionality of Dyld for our
    own purposes.

    It isn't meant to be a faithful recreation of dyld so to speak, it just does things dyld also does, kinda.

    """

    @classmethod
    def load(cls, macho_slice, load_symtab=True, load_imports=True, load_exports=True):
        """
        Take a slice of a macho file and process it using the dyld functions

        :param load_binding: Load Binding Info?
        :param load_symtab: Load Symbol Table?
        :param macho_slice: Slice to load. If your library is not fat, that'll be MachOFile.slices[0]
        :type macho_slice: Slice
        :return: Processed Library object
        :rtype: Library
        """
        log.info("Loading Library")
        library = Library(macho_slice)

        log.info("Processing Load Commands")
        Dyld._parse_load_commands(library, load_symtab, load_imports, load_exports)
        return library

    @classmethod
    def _parse_load_commands(cls, library, load_symtab=True, load_imports=True, load_exports=True):
        # noinspection PyUnusedLocal
        fixups = None
        log.info(f'registered {len(library.macho_header.load_commands)} Load Commands')
        for cmd in library.macho_header.load_commands:
            if isinstance(cmd, segment_command_64):
                log.debug("Loading segment_command_64")
                segment = Segment(library, cmd)

                log.info(f'Loaded Segment {segment.name}')
                library.vm.add_segment(segment)
                library.segments[segment.name] = segment

                log.debug(f'Added {segment.name} to VM Map')

            elif isinstance(cmd, dyld_info_command):
                library.info = cmd
                if load_imports:
                    log.info("Loading Binding Info")
                    library.binding_table = BindingTable(library, cmd.bind_off, cmd.bind_size)
                    library.weak_binding_table = BindingTable(library, cmd.weak_bind_off, cmd.weak_bind_size)
                    library.lazy_binding_table = BindingTable(library, cmd.lazy_bind_off, cmd.lazy_bind_size)
                if load_exports:
                    log.info("Loading Export Trie")
                    library.exports = ExportTrie.from_bytes(library, cmd.export_off, cmd.export_size)

            elif LOAD_COMMAND(cmd.cmd) == LOAD_COMMAND.LC_DYLD_EXPORTS_TRIE:
                log.info("Loading Export Trie")
                library.exports = ExportTrie.from_bytes(library, cmd.dataoff, cmd.datasize)

            elif LOAD_COMMAND(cmd.cmd) == LOAD_COMMAND.LC_DYLD_CHAINED_FIXUPS:
                log.warning("library uses LC_DYLD_CHAINED_FIXUPS; This is not yet supported in ktool, off-image symbol resolution (superclasses, etc) will not work")
                # fixups = ChainedFixups(library, cmd.dataoff, cmd.datasize)
                pass

            elif isinstance(cmd, symtab_command):
                if load_symtab:
                    log.info("Loading Symbol Table")
                    library.symbol_table = SymbolTable(library, cmd)

            elif isinstance(cmd, uuid_command):
                library.uuid = cmd.uuid.to_bytes(16, "little")
                log.info(f'Library UUID: {library.uuid}')

            elif isinstance(cmd, sub_client_command):
                string = library.get_cstr_at(cmd.off + cmd.offset)
                library.allowed_clients.append(string)
                log.debug(f'Loaded Subclient "{string}"')

            elif isinstance(cmd, rpath_command):
                string = library.get_cstr_at(cmd.off + cmd.path)
                library.rpath = string
                log.info(f'Library Resource Path: {string}')

            elif isinstance(cmd, build_version_command):
                library.platform = PlatformType(cmd.platform)
                library.minos = os_version(x=library.get_int_at(cmd.off + 14, 2), y=library.get_int_at(cmd.off + 13, 1),
                                           z=library.get_int_at(cmd.off + 12, 1))
                library.sdk_version = os_version(x=library.get_int_at(cmd.off + 18, 2),
                                                 y=library.get_int_at(cmd.off + 17, 1),
                                                 z=library.get_int_at(cmd.off + 16, 1))
                log.info(f'Loaded platform {library.platform.name} | '
                          f'Minimum OS {library.minos.x}.{library.minos.y}'
                          f'.{library.minos.z} | SDK Version {library.sdk_version.x}'
                          f'.{library.sdk_version.y}.{library.sdk_version.z}')

            elif isinstance(cmd, version_min_command):
                # Only override this if it wasn't set by build_version
                if library.platform == PlatformType.UNK:
                    if LOAD_COMMAND(cmd.cmd) == LOAD_COMMAND.VERSION_MIN_MACOSX:
                        library.platform = PlatformType.MACOS
                    elif LOAD_COMMAND(cmd.cmd) == LOAD_COMMAND.VERSION_MIN_IPHONEOS:
                        library.platform = PlatformType.IOS
                    elif LOAD_COMMAND(cmd.cmd) == LOAD_COMMAND.VERSION_MIN_TVOS:
                        library.platform = PlatformType.TVOS
                    elif LOAD_COMMAND(cmd.cmd) == LOAD_COMMAND.VERSION_MIN_WATCHOS:
                        library.platform = PlatformType.WATCHOS

                    library.minos = os_version(x=library.get_int_at(cmd.off + 10, 2),
                                               y=library.get_int_at(cmd.off + 9, 1),
                                               z=library.get_int_at(cmd.off + 8, 1))

            elif isinstance(cmd, dylib_command):
                if cmd.cmd == 0xD:  # local
                    library.dylib = ExternalDylib(library, cmd)
                    log.info(f'Loaded local dylib_command with install_name {library.dylib.install_name}')
                else:
                    external_dylib = ExternalDylib(library, cmd)
                    library.linked.append(external_dylib)
                    log.info(f'Loaded linked dylib_command with install name {external_dylib.install_name}')

        if library.dylib is not None:
            library.name = library.dylib.install_name.split('/')[-1]
        else:
            library.name = ""


class Library:
    """
    This class represents the Mach-O Binary as a whole.

    It's the root object in the massive tree of information we're going to build up about the binary

    This is an abstracted version, other classes will handle the raw struct interaction;
        here, we facilitate that interaction within those classes and generate our abstract representation

    This class on its own does not handle populating its fields.
    The Dyld class set is responsible for loading in and processing the raw values to it.
    """

    def __init__(self, macho_slice):
        """
        Create a MachO Library

        :param macho_slice: MachO Slice being processed
        :type macho_slice: MachO Slice
        """
        self.macho_header = LibraryHeader.from_bytes(macho_slice=macho_slice)

        self.slice = macho_slice

        self.linked = []
        self.name = ""
        self.segments = {}

        log.debug("Initializing VM Map")
        self.vm = _VirtualMemoryMap(macho_slice)

        self.info = None
        self.dylib = None
        self.uuid = None

        self.platform = PlatformType.UNK

        self.allowed_clients = []

        self.rpath = None

        self.minos = os_version(0, 0, 0)
        self.sdk_version = os_version(0, 0, 0)

        self.binding_table = None
        self.weak_binding_table = None
        self.lazy_binding_table = None
        self.exports = None

        self.symbol_table = None

    def get_int_at(self, offset: int, length: int, vm=False, section_name=None):
        """
        Get a sequence of bytes (as an int) from a location

        :param offset: Offset within the library
        :param length: Amount of bytes to get
        :param vm: Is `offset` a VM address
        :param section_name: Section Name if vm==True (improves translation time slightly)
        :return: `length` Bytes at `offset`
        """
        if vm:
            offset = self.vm.get_file_address(offset, section_name)
        return self.slice.get_int_at(offset, length)

    def get_bytes_at(self, offset: int, length: int, vm=False, section_name=None):
        """
        Get a sequence of bytes from a location

        :param offset: Offset within the library
        :param length: Amount of bytes to get
        :param vm: Is `offset` a VM address
        :param section_name: Section Name if vm==True (improves translation time slightly)
        :return: `length` Bytes at `offset`
        """
        if vm:
            offset = self.vm.get_file_address(offset, section_name)
        return self.slice.get_bytes_at(offset, length)

    def load_struct(self, address: int, struct_type, vm=False, section_name=None, endian="little"):
        """
        Load a struct (struct_type_t) from a location and return the processed object

        :param address: Address to load struct from
        :param struct_type: type of struct (e.g. dyld_header)
        :param vm:  Is `address` a VM address?
        :param section_name: if `vm==True`, the section name (slightly improves translation speed)
        :param endian: Endianness of bytes to read.
        :return: Loaded struct
        """
        if vm:
            address = self.vm.get_file_address(address, section_name)
        return self.slice.load_struct(address, struct_type, endian)

    def get_str_at(self, address: int, count: int, vm=False, section_name=None):
        """
        Get string with set length from location (to be used essentially only for loading segment names)

        :param address: Address of string start
        :param count: Length of string
        :param vm: Is `address` a VM address?
        :param section_name: if `vm==True`, the section name (unused here, really)
        :return: The loaded string.
        """
        if vm:
            address = self.vm.get_file_address(address, section_name)
        return self.slice.get_str_at(address, count)

    def get_cstr_at(self, address: int, limit: int = 0, vm=False, section_name=None):
        """
        Load a C style string from a location, stopping once a null byte is encountered.

        :param address: Address to load string from
        :param limit: Limit of the length of bytes, 0 = unlimited
        :param vm: Is `address` a VM address?
        :param section_name: if `vm==True`, the section name (vastly improves VM lookup time)
        :return: The loaded C string
        """
        if vm:
            address = self.vm.get_file_address(address, section_name)
        return self.slice.get_cstr_at(address, limit)

    def decode_uleb128(self, readHead: int):
        """
        Decode a uleb128 integer from a location

        :param readHead: Start location
        :return: (end location, value)
        """
        return self.slice.decode_uleb128(readHead)

    def rm_load_command(self, index):
        b_load_cmd = self.macho_header.load_commands.pop(index)

        off = b_load_cmd.off + b_load_cmd.cmdsize
        after_bytes = self.macho_header.raw_bytes()[off:self.macho_header.dyld_header.loadsize + 32]

        self.slice.patch(b_load_cmd.off, after_bytes)
        self.slice.patch(self.macho_header.dyld_header.loadsize + 32 - b_load_cmd.cmdsize, b'\x00' * b_load_cmd.cmdsize)

        dyld_head = self.macho_header.dyld_header
        dyld_head.loadcnt -= 1
        dyld_head.loadsize -= b_load_cmd.cmdsize

        self.slice.patch(self.macho_header.dyld_header.off, dyld_head.raw)

    def insert_lc(self, lc, fields, index=-1):
        lc_type = LOAD_COMMAND_MAP[lc]

        load_cmd = Struct.create_with_values(lc_type, [lc.value, lc_type.SIZE] + fields)

        off = dyld_header.SIZE
        off += self.macho_header.dyld_header.loadsize
        raw = load_cmd.raw
        size = len(load_cmd.raw)

        if index != -1:
            b_load_cmd = self.macho_header.load_commands[index - 1]
            off = b_load_cmd.off + b_load_cmd.cmdsize
            after_bytes = self.macho_header.raw_bytes()[off:self.macho_header.dyld_header.loadsize + 32]
            self.slice.patch(off, raw)
            self.slice.patch(off + size, after_bytes)
        else:
            self.slice.patch(off, raw)

        self.macho_header.load_commands.append(load_cmd)

        dyld_head = self.macho_header.dyld_header
        dyld_head.loadcnt += 1
        dyld_head.loadsize += size

        self.slice.patch(self.macho_header.dyld_header.off, dyld_head.raw)

    def insert_lc_with_suf(self, lc, fields, suffix, index=-1):
        lc_type = LOAD_COMMAND_MAP[lc]

        load_cmd = Struct.create_with_values(lc_type, [lc.value, lc_type.SIZE] + fields)
        # log.debug(f'Fabricated Load Command {str(load_cmd)}')

        encoded = suffix.encode('utf-8') + b'\x00'

        cmdsize = lc_type.SIZE
        cmdsize += len(encoded)
        cmdsize = 0x8 * math.ceil(cmdsize / 0x8)
        log.debug(f'Computed Struct Size of {cmdsize}')

        load_cmd.cmdsize = cmdsize

        off = dyld_header.SIZE
        off += self.macho_header.dyld_header.loadsize
        raw = load_cmd.raw + encoded + (b'\x00' * (cmdsize - (lc_type.SIZE + len(encoded))))
        log.debug(f'Padding Size {(cmdsize - (lc_type.SIZE + len(encoded)))}')
        size = len(raw)

        if index != -1:
            b_load_cmd = self.macho_header.load_commands[index - 1]
            off = b_load_cmd.off + b_load_cmd.cmdsize
            after_bytes = self.macho_header.raw_bytes()[off:self.macho_header.dyld_header.loadsize + 32]
            self.slice.patch(off, raw)
            self.slice.patch(off + size, after_bytes)
        else:
            self.slice.patch(off, raw)

        self.macho_header.load_commands.append(load_cmd)

        dyld_head = self.macho_header.dyld_header
        dyld_head.loadcnt += 1
        dyld_head.loadsize += size

        self.slice.patch(self.macho_header.dyld_header.off, dyld_head.raw)


class LibraryHeader(Constructable):
    """
    This class represents the Mach-O Header
    It contains the basic header info along with all load commands within it.

    It doesn't handle complex abstraction logic, it simply loads in the load commands as their raw structs
    """

    @staticmethod
    def from_bytes(macho_slice):

        library_header = LibraryHeader()

        offset = 0
        header: dyld_header = macho_slice.load_struct(offset, dyld_header)
        raw = header.raw

        library_header.filetype = MH_FILETYPE(header.filetype)

        for flag in MH_FLAGS:
            if header.flags & flag.value:
                library_header.flags.append(flag)

        offset += header.SIZE

        load_commands = []

        for i in range(1, header.loadcnt + 1):
            cmd = macho_slice.get_int_at(offset, 4)
            cmd_size = macho_slice.get_int_at(offset+4, 4)

            cmd_raw = macho_slice.get_bytes_at(offset, cmd_size)
            try:
                load_cmd = Struct.create_with_bytes(LOAD_COMMAND_MAP[LOAD_COMMAND(cmd)], cmd_raw)
                load_cmd.off = offset
            except ValueError:
                unk_lc = macho_slice.load_struct(offset, unk_command)
                load_cmd = unk_lc
            except KeyError:
                unk_lc = macho_slice.load_struct(offset, unk_command)
                load_cmd = unk_lc

            load_commands.append(load_cmd)
            raw += cmd_raw
            offset += load_cmd.cmdsize

        library_header.raw = raw
        library_header.dyld_header = header
        library_header.load_commands = load_commands

        return library_header

    @staticmethod
    def from_values(*args, **kwargs):
        pass

    def __init__(self):
        self.dyld_header = None
        self.filetype = MH_FILETYPE(0)
        self.flags = []
        self.load_commands = []
        self.raw = bytearray()

    def raw_bytes(self):
        return self.raw


class ExternalDylib:
    def __init__(self, source_library, cmd):
        self.cmd = cmd
        self.source_library = source_library
        self.install_name = self._get_name(cmd)
        self.weak = cmd.cmd == 0x18 | 0x80000000
        self.local = cmd.cmd == 0xD

    def _get_name(self, cmd):
        read_address = cmd.off + dylib_command.SIZE
        return self.source_library.get_cstr_at(read_address)


os_version = namedtuple("os_version", ["x", "y", "z"])


class PlatformType(Enum):
    MACOS = 1
    IOS = 2
    TVOS = 3
    WATCHOS = 4
    BRIDGE_OS = 5
    MAC_CATALYST = 6
    IOS_SIMULATOR = 7
    TVOS_SIMULATOR = 8
    WATCHOS_SIMULATOR = 9
    DRIVER_KIT = 10
    UNK = 64


class ToolType(Enum):
    CLANG = 1
    SWIFT = 2
    LD = 3


class SymbolType(Enum):
    CLASS = 0
    METACLASS = 1
    IVAR = 2
    FUNC = 3
    UNK = 4


class Symbol:
    """
    This class can represent several types of symbols.

    It can represent an external or internal symbol declaration and is used for both across the library

    .external is a BOOL that can be used to check whether it's an external or internal declaration

    .fullname contains the full name of the symbol (e.g. _OBJC_CLASS_$_MyDumbClassnameHere)

    .name contains the (somewhat) processed name of the symbol (e.g. _MyDumbClassnameHere for an @interface
    MyDumbClassnameHere)

    .type contains a SymbolType if it was able to figure one out

    .addr contains the address of the symbol in the image

    """

    def __init__(self, lib, cmd=None, entry=None, fullname=None, ordinal=None, addr=None):
        if fullname:
            self.fullname = fullname
        else:
            self.fullname = lib.get_cstr_at(entry.str_index + cmd.stroff)
        if '_$_' in self.fullname:
            if self.fullname.startswith('_OBJC_CLASS_$'):
                self.type = SymbolType.CLASS
            elif self.fullname.startswith('_OBJC_METACLASS_$'):
                self.type = SymbolType.METACLASS
            elif self.fullname.startswith('_OBJC_IVAR_$'):
                self.type = SymbolType.IVAR
            else:
                self.type = SymbolType.UNK
            self.name = self.fullname.split('$')[1]
        else:
            self.name = self.fullname
            self.type = SymbolType.FUNC
        if entry:
            self.external = False
            self.addr = entry.value
        else:
            self.external = True
            self.addr = addr
        self.entry = entry
        self.ordinal = ordinal


class SymbolTable:
    """
    This class represents the symbol table declared in the MachO File

    .table contains the symbol table

    .ext contains exported symbols, i think?

    This class is incomplete

    """

    def __init__(self, library, cmd: symtab_command):
        self.library = library
        self.cmd = cmd
        self.ext = []
        self.table = self._load_symbol_table()

    def _load_symbol_table(self):
        symbol_table = []
        read_address = self.cmd.symoff
        for i in range(0, self.cmd.nsyms):
            symbol_table.append(self.library.load_struct(read_address + symtab_entry.SIZE * i, symtab_entry))

        table = []
        for sym in symbol_table:
            symbol = Symbol(self.library, self.cmd, sym)
            # log.debug(f'Symbol Table: Loaded symbol:{symbol.name} ordinal:{symbol.ordinal} type:{symbol.type}')
            table.append(symbol)
            if sym.type == 0xf:
                self.ext.append(symbol)
        return table


export_node = namedtuple("export_node", ['text', 'offset'])


class ExportTrie(Constructable):
    @staticmethod
    def from_bytes(library, export_start, export_size):
        trie = ExportTrie()

        endpoint = export_start + export_size
        nodes = ExportTrie.read_node(library, export_start, '', export_start, endpoint)
        symbols = []

        for node in nodes:
            symbols.append(Symbol(library, fullname=node.text, addr=node.offset))

        trie.nodes = nodes
        trie.symbols = symbols
        trie.raw = library.get_bytes_at(export_start, export_size)

        return trie

    @staticmethod
    def from_values(*args, **kwargs):
        pass

    def raw_bytes(self):
        return self.raw

    def __init__(self):
        self.raw = bytearray()
        self.nodes = []
        self.symbols = []

    @classmethod
    def read_node(cls, library, trie_start, string, cursor, endpoint):

        if cursor > endpoint:
            macho_is_malformed()

        start = cursor
        byte = library.get_int_at(cursor, 1)
        results = []
        log.debug_tm(f'@ {hex(start)} node: {hex(byte)} current_symbol: {string}')
        if byte == 0:
            cursor += 1
            branches = library.get_int_at(cursor, 1)
            log.debug_tm(f'BRAN {branches}')
            for i in range(0, branches):
                if i == 0:
                    cursor += 1
                proc_str = library.get_cstr_at(cursor)
                cursor += len(proc_str) + 1
                offset, cursor = library.decode_uleb128(cursor)
                log.debug_tm(f'({i}) string: {string + proc_str} next_node: {hex(trie_start + offset)}')
                results += ExportTrie.read_node(library, trie_start, string + proc_str, trie_start + offset, endpoint)
        else:
            log.debug_tm(f'TERM: 0')
            size, cursor = library.decode_uleb128(cursor)
            flags = library.get_int_at(cursor, 1)
            cursor += 1
            offset, cursor = library.decode_uleb128(cursor)
            results.append(export_node(string, offset))

        return results


action = namedtuple("action", ["vmaddr", "libname", "item"])
record = namedtuple("record", [
    "off",
    "seg_index",
    "seg_offset",
    "lib_ordinal",
    "type",
    "flags",
    "name",
    "addend",
    "special_dylib"
])


class BindingTable:
    """
    The binding table contains a ton of information related to the binding info in the library

    .lookup_table - Contains a map of address -> Symbol declarations which should be used for processing off-image
    symbol decorations

    .symbol_table - Contains a full list of symbols declared in the binding info. Avoid iterating through this for
    speed purposes.

    .actions - contains a list of, you guessed it, actions.

    .import_stack - contains a fairly raw unprocessed list of binding info commands

    """

    def __init__(self, library, table_start, table_size):
        """
        Pass a library to be processed

        :param library: Library to be processed
        :type library: Library
        """
        self.library = library
        self.import_stack = self._load_binding_info(table_start, table_size)
        self.actions = self._create_action_list()
        self.lookup_table = {}
        self.link_table = {}
        self.symbol_table = self._load_symbol_table()

    def _load_symbol_table(self):
        table = []
        for act in self.actions:
            if act.item:
                sym = Symbol(self.library, fullname=act.item, ordinal=act.libname, addr=act.vmaddr)
                # log.debug(f'Binding info: Loaded symbol:{act.item} ordinal:{act.libname} addr:{act.vmaddr}')
                table.append(sym)
                self.lookup_table[act.vmaddr] = sym
        return table

    def _create_action_list(self):
        actions = []
        for bind_command in self.import_stack:
            segment = list(self.library.segments.values())[bind_command.seg_index]
            vm_address = segment.vm_address + bind_command.seg_offset
            try:
                lib = self.library.linked[bind_command.lib_ordinal - 1].install_name
            except IndexError:
                # log.debug(f'Binding Info: {bind_command.lib_ordinal} Ordinal wasn't found, Something is wrong')
                lib = str(bind_command.lib_ordinal)
            item = bind_command.name
            actions.append(action(vm_address & 0xFFFFFFFFF, lib, item))
        return actions

    def _load_binding_info(self, table_start, table_size):
        read_address = table_start
        import_stack = []
        while True:
            if read_address - table_size >= table_start:
                break
            seg_index = 0x0
            seg_offset = 0x0
            lib_ordinal = 0x0
            btype = 0x0
            flags = 0x0
            name = ""
            addend = 0x0
            special_dylib = 0x0
            while True:
                # There are 0xc opcodes total
                # Bitmask opcode byte with 0xF0 to get opcode, 0xF to get value
                binding_opcode = self.library.get_int_at(read_address, 1) & 0xF0
                value = self.library.get_int_at(read_address, 1) & 0x0F
                log.debug_tm(f'{BINDING_OPCODE(binding_opcode).name}: {hex(value)}')
                cmd_start_addr = read_address
                read_address += 1

                if binding_opcode == BINDING_OPCODE.DONE:
                    import_stack.append(
                        record(cmd_start_addr, seg_index, seg_offset, lib_ordinal, btype, flags, name, addend,
                               special_dylib))
                    break

                elif binding_opcode == BINDING_OPCODE.SET_DYLIB_ORDINAL_IMM:
                    lib_ordinal = value

                elif binding_opcode == BINDING_OPCODE.SET_DYLIB_ORDINAL_ULEB:
                    lib_ordinal, read_address = self.library.decode_uleb128(read_address)

                elif binding_opcode == BINDING_OPCODE.SET_DYLIB_SPECIAL_IMM:
                    special_dylib = 0x1
                    lib_ordinal = value

                elif binding_opcode == BINDING_OPCODE.SET_SYMBOL_TRAILING_FLAGS_IMM:
                    flags = value
                    name = self.library.get_cstr_at(read_address)
                    read_address += len(name) + 1

                elif binding_opcode == BINDING_OPCODE.SET_TYPE_IMM:
                    btype = value

                elif binding_opcode == BINDING_OPCODE.SET_ADDEND_SLEB:
                    addend, read_address = self.library.decode_uleb128(read_address)

                elif binding_opcode == BINDING_OPCODE.SET_SEGMENT_AND_OFFSET_ULEB:
                    seg_index = value
                    seg_offset, read_address = self.library.decode_uleb128(read_address)

                elif binding_opcode == BINDING_OPCODE.ADD_ADDR_ULEB:
                    o, read_address = self.library.decode_uleb128(read_address)
                    seg_offset += o

                elif binding_opcode == BINDING_OPCODE.DO_BIND_ADD_ADDR_ULEB:
                    import_stack.append(
                        record(cmd_start_addr, seg_index, seg_offset, lib_ordinal, btype, flags, name, addend,
                               special_dylib))
                    seg_offset += 8
                    o, read_address = self.library.decode_uleb128(read_address)
                    seg_offset += o

                elif binding_opcode == BINDING_OPCODE.DO_BIND_ADD_ADDR_IMM_SCALED:
                    import_stack.append(
                        record(cmd_start_addr, seg_index, seg_offset, lib_ordinal, btype, flags, name, addend,
                               special_dylib))
                    seg_offset = seg_offset + (value * 8) + 8

                elif binding_opcode == BINDING_OPCODE.DO_BIND_ULEB_TIMES_SKIPPING_ULEB:
                    count, read_address = self.library.decode_uleb128(read_address)
                    skip, read_address = self.library.decode_uleb128(read_address)

                    for i in range(0, count):
                        import_stack.append(
                            record(cmd_start_addr, seg_index, seg_offset, lib_ordinal, btype, flags, name, addend,
                                   special_dylib))
                        seg_offset += skip + 8

                elif binding_opcode == BINDING_OPCODE.DO_BIND:
                    import_stack.append(
                        record(cmd_start_addr, seg_index, seg_offset, lib_ordinal, btype, flags, name, addend,
                               special_dylib))
                    seg_offset += 8

        return import_stack



#
#  ktool | kmacho
#  structs.py
#
#  Custom Struct implementation reflecting behavior of named tuples while also handling behind-the-scenes packing/unpacking.
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2021.
#

class Struct:
    """
    Custom namedtuple-esque Struct representation. Can be unpacked from bytes or manually created with existing
        field values

    Subclassed and passed `fields` and `sizes` values.

    Fields are exposed as read-write attributes, and when written to, will update the backend
        byte representation of the struct, accessible via the .raw attribute

    """

    @staticmethod
    def create_with_bytes(struct_class, raw, byte_order="little"):
        """
        Unpack a struct from raw bytes

        :param struct_class: Struct subclass
        :param raw: Bytes
        :param byte_order: Little/Big Endian Struct Unpacking
        :return: struct_class Instance
        """
        instance: Struct = struct_class(byte_order)
        current_off = 0
        raw = bytearray(raw)

        for field in instance._field_list:
            size = instance._field_sizes[field]
            data = raw[current_off:current_off + size]

            instance._fields[field] = int.from_bytes(data, byte_order)

            instance.raw += data
            current_off += size

        instance.initialized = True
        return instance

    @staticmethod
    def create_with_values(struct_class, values, byte_order="little"):
        """
        Pack/Create a struct given field values

        :param byte_order:
        :param struct_class: Struct subclass
        :param values: List of values
        :return: struct_class Instance
        """

        instance: Struct = struct_class(byte_order)

        for i, field in enumerate(instance._field_list):
            instance._fields[field] = values[i]

        instance._rebuild_raw()
        instance.initialized = True
        return instance

    def typename(self):
        return self.__class__.__name__

    def desc(self):
        return ""

    def __str__(self):
        text = f'{self.__class__.__name__}('
        for field in self._field_list:
            text += f'{field}={hex(self._fields[field])}, '
        return text[:-2] + ')'

    def __init__(self, fields=None, sizes=None, byte_order="little"):
        if sizes is None:
            raise AssertionError(
                "Do not use the bare Struct class; it must be implemented in an actual type; Missing Sizes")

        if fields is None:
            raise AssertionError(
                "Do not use the bare Struct class; it must be implemented in an actual type; Missing Fields")

        # Declare this first to avoid a recursion error
        self._fields = {}

        self.byte_order = byte_order
        self.initialized = False

        self._field_list = fields
        self._field_sizes = {}
        self._sizeof = sum(sizes)

        for index, i in enumerate(fields):
            self._fields[i] = 0
            self._field_sizes[i] = sizes[index]

        self.off = 0
        self.raw = bytearray(b'')

    def _rebuild_raw(self):
        raw = bytearray()
        for field in self._field_list:
            size = self._field_sizes[field]

            field_dat = self._fields[field]

            data = None

            if isinstance(field_dat, int):
                data = field_dat.to_bytes(size, byteorder=self.byte_order)
            elif isinstance(field_dat, bytearray) or isinstance(field_dat, bytes):
                data = self._fields[field]

            assert data is not None

            raw += bytearray(data)

        assert len(raw) == self._sizeof

        self.raw = raw

    def __len__(self):
        return self._sizeof

    def __getattr__(self, item):
        if item == '_fields':
            return {}
        if item in self._fields:
            return self._fields[item]
        raise AttributeError(f'{item} not in struct or internal properties')

    def __setattr__(self, key, value):
        if key in self._fields:
            self._fields[key] = value
            if self.initialized:
                self._rebuild_raw()
        else:
            super().__setattr__(key, value)


class fat_header(Struct):
    """
    First 8 Bytes of a FAT MachO File

    Attributes:
        self.magic: FAT MachO Magic

        self.nfat_archs: Number of Fat Arch entries after these bytes
    """
    _FIELDNAMES = ['magic', 'nfat_archs']
    _SIZES = [4, 4]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class fat_arch(Struct):
    """
    Struct representing a slice in a FAT MachO

    Attribs:
        cpu_type:
    """
    _FIELDNAMES = ['cpu_type', 'cpu_subtype', 'offset', 'size', 'align']
    _SIZES = [4, 4, 4, 4, 4]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class dyld_header(Struct):
    _FIELDNAMES = ['magic', 'cpu_type', 'cpu_subtype', 'filetype', 'loadcnt', 'loadsize', 'flags', 'void']
    _SIZES = [4, 4, 4, 4, 4, 4, 4, 4]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class unk_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize']
    _SIZES = [4, 4]
    SIZE = 8

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class segment_command_64(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'segname', 'vmaddr', 'vmsize', 'fileoff', 'filesize', 'maxprot',
                   'initprot', 'nsects', 'flags']
    _SIZES = [4, 4, 16, 8, 8, 8, 8, 4, 4, 4, 4]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class section_64(Struct):
    _FIELDNAMES = ['sectname', 'segname', 'addr', 'size', 'offset', 'align', 'reloff',
                   'nreloc', 'flags', 'reserved1', 'reserved2', 'reserved3']
    _SIZES = [16, 16, 8, 8, 4, 4, 4, 4, 4, 4, 4, 4]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class symtab_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'symoff', 'nsyms', 'stroff', 'strsize']
    _SIZES = [4, 4, 4, 4, 4, 4]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class dysymtab_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'ilocalsym', 'nlocalsym', 'iextdefsym', 'nextdefsym',
                   'iundefsym', 'nundefsym', 'tocoff', 'ntoc', 'modtaboff', 'nmodtab',
                   'extrefsymoff', 'nextrefsyms', 'indirectsymoff', 'nindirectsyms',
                   'extreloff', 'nextrel', 'locreloff', 'nlocrel']
    _SIZES = [4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class dylib(Struct):
    _FIELDNAMES = ['name', 'timestamp', 'current_version', 'compatibility_version']
    _SIZES = [4, 4, 4, 4]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class dylib_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'dylib']
    _SIZES = [4, 4, 16]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class dylinker_command(Struct):
    _FIELDNAMES = ["cmd", "cmdsize", "name"]
    _SIZES = [4, 4, 4]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class sub_client_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'offset']
    _SIZES = [4, 4, 4]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class uuid_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'uuid']
    _SIZES = [4, 4, 16]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class build_version_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'platform', 'minos', 'sdk', 'ntools']
    _SIZES = [4, 4, 4, 4, 4, 4]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class entry_point_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'entryoff', 'stacksize']
    _SIZES = [4, 4, 8, 8]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class rpath_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'path']
    _SIZES = [4, 4, 4]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class source_version_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'version']
    _SIZES = [4, 4, 8]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class linkedit_data_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'dataoff', 'datasize']
    _SIZES = [4, 4, 4, 4]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class dyld_info_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'rebase_off', 'rebase_size', 'bind_off', 'bind_size',
                   'weak_bind_off', 'weak_bind_size', 'lazy_bind_off', 'lazy_bind_size',
                   'export_off', 'export_size']
    _SIZES = [4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class symtab_entry(Struct):
    _FIELDNAMES = ["str_index", "type", "sect_index", "desc", "value"]
    _SIZES = [4, 1, 1, 2, 8]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class version_min_command(Struct):
    _FIELDNAMES = ["cmd", "cmdsize", "version", "reserved"]
    _SIZES = [4, 4, 4, 4]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)



#
#  ktool | kmacho
#  mach_header.py
#
#
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2021.
#


from enum import IntEnum

from .structs import *

MH_MAGIC = 0xFEEDFACE
MH_CIGAM = 0xCEFAEDFE
MH_MAGIC_64 = 0xFEEDFACF
MH_CIGAM_64 = 0xCFFAEDFE
FAT_MAGIC = 0xCAFEBABE
FAT_CIGAM = 0xBEBAFECA


class MH_FLAGS(IntEnum):
    NOUNDEFS = 0x1
    INCRLINK = 0x2
    DYLDLINK = 0x4
    BINDATLOAD = 0x8
    PREBOUND = 0x10
    SPLIT_SEGS = 0x20
    LAZY_INIT = 0x40
    TWOLEVEL = 0x80
    FORCE_FLAT = 0x100
    NOMULTIEFS = 0x200
    NOFIXPREBINDING = 0x400
    PREBINDABLE = 0x800
    ALLMODSBOUND = 0x1000
    SUBSECTIONS_VIA_SYMBOLS = 0x2000
    CANONICAL = 0x4000
    WEAK_DEFINES = 0x8000
    BINDS_TO_WEAK = 0x10000
    ALLOW_STACK_EXECUTION = 0x20000
    ROOT_SAFE = 0x40000
    SETUID_SAFE = 0x80000
    NO_REEXPORTED_DYLIBS = 0x100000
    PIE = 0x200000
    DEAD_STRIPPABLE_DYLIB = 0x400000
    HAS_TLV_DESCRIPTORS = 0x800000
    NO_HEAP_EXECUTION = 0x1000000
    APP_EXTENSION_SAFE = 0x02000000
    NLIST_OUTOFSYNC_WITH_DYLDINFO = 0x04000000
    SIM_SUPPORT = 0x08000000


class MH_FILETYPE(IntEnum):
    UNK = 0
    OBJECT = 0x1
    EXECUTE = 0x2
    FVMLIB = 0x3
    CORE = 0x4
    PRELOAD = 0x5
    DYLIB = 0x6
    DYLINKER = 0x7
    BUNDLE = 0x8
    DYLIB_STUB = 0x9
    DSYM = 0xA
    KEXT_BUNDLE = 0xB


LC_REQ_DYLD = 0x80000000


class LOAD_COMMAND(IntEnum):
    SEGMENT = 0x1
    SYMTAB = 0x2
    SYMSEG = 0x3
    THREAD = 0x4
    UNIXTHREAD = 0x5
    LOADFVMLIB = 0x6
    IDFVMLIB = 0x7
    IDENT = 0x8
    FVMFILE = 0x9
    PREPAGE = 0xA
    DYSYMTAB = 0xB
    LOAD_DYLIB = 0xC
    ID_DYLIB = 0xD
    LOAD_DYLINKER = 0xE
    ID_DYLINKER = 0xF
    PREBOUND_DYLIB = 0x10
    ROUTINES = 0x11
    SUB_FRAMEWORK = 0x12
    SUB_UMBRELLA = 0x13
    SUB_CLIENT = 0x14
    SUB_LIBRARY = 0x15
    TWOLEVEL_HINTS = 0x16
    PREBIND_CKSUM = 0x17
    LOAD_WEAK_DYLIB = 0x18 | LC_REQ_DYLD
    SEGMENT_64 = 0x19
    ROUTINES_64 = 0x1a
    UUID = 0x1b
    RPATH = 0x1C | LC_REQ_DYLD
    CODE_SIGNATURE = 0x1D
    SEGMENT_SPLIT_INFO = 0x1E
    REEXPORT_DYLIB = 0x1F | LC_REQ_DYLD
    LAZY_LOAD_DYLIB = 0x20
    ENCRYPTION_INFO = 0x21
    DYLD_INFO = 0x22
    DYLD_INFO_ONLY = 0x22 | LC_REQ_DYLD
    LOAD_UPWARD_DYLIB = 0x23 | LC_REQ_DYLD
    VERSION_MIN_MACOSX = 0x24
    VERSION_MIN_IPHONEOS = 0x25
    FUNCTION_STARTS = 0x26
    DYLD_ENVIRONMENT = 0x27
    MAIN = 0x28 | LC_REQ_DYLD
    DATA_IN_CODE = 0x29
    SOURCE_VERSION = 0x2A
    DYLIB_CODE_SIGN_DRS = 0x2B
    ENCRYPTION_INFO_64 = 0x2C
    LINKER_OPTION = 0x2D
    LINKER_OPTIMIZATION_HINT = 0x2E
    VERSION_MIN_TVOS = 0x2F
    VERSION_MIN_WATCHOS = 0x30
    NOTE = 0x31
    BUILD_VERSION = 0x32
    LC_DYLD_EXPORTS_TRIE = 0x33 | LC_REQ_DYLD
    LC_DYLD_CHAINED_FIXUPS = 0x34 | LC_REQ_DYLD


LOAD_COMMAND_MAP = {
    LOAD_COMMAND.SYMTAB: symtab_command,
    LOAD_COMMAND.DYSYMTAB: dysymtab_command,
    LOAD_COMMAND.LOAD_DYLIB: dylib_command,
    LOAD_COMMAND.ID_DYLIB: dylib_command,
    LOAD_COMMAND.LOAD_DYLINKER: dylinker_command,
    LOAD_COMMAND.SUB_CLIENT: sub_client_command,
    LOAD_COMMAND.LOAD_WEAK_DYLIB: dylib_command,
    LOAD_COMMAND.SEGMENT_64: segment_command_64,
    LOAD_COMMAND.UUID: uuid_command,
    LOAD_COMMAND.CODE_SIGNATURE: linkedit_data_command,
    LOAD_COMMAND.SEGMENT_SPLIT_INFO: linkedit_data_command,
    LOAD_COMMAND.SOURCE_VERSION: source_version_command,
    LOAD_COMMAND.DYLD_INFO_ONLY: dyld_info_command,
    LOAD_COMMAND.FUNCTION_STARTS: linkedit_data_command,
    LOAD_COMMAND.DATA_IN_CODE: linkedit_data_command,
    LOAD_COMMAND.BUILD_VERSION: build_version_command,
    LOAD_COMMAND.MAIN: entry_point_command,
    LOAD_COMMAND.RPATH: rpath_command,
    LOAD_COMMAND.VERSION_MIN_MACOSX: version_min_command,
    LOAD_COMMAND.VERSION_MIN_IPHONEOS: version_min_command,
    LOAD_COMMAND.VERSION_MIN_TVOS: version_min_command,
    LOAD_COMMAND.VERSION_MIN_WATCHOS: version_min_command,
    LOAD_COMMAND.LC_DYLD_EXPORTS_TRIE: linkedit_data_command,
    LOAD_COMMAND.LC_DYLD_CHAINED_FIXUPS: linkedit_data_command
}


class S_FLAGS_MASKS(IntEnum):
    SECTION_TYPE = 0x000000ff
    SECTION_ATTRIBUTES = 0xffffff00
    SECTION_ATTRIBUTES_USR = 0xff000000
    SECTION_ATTRIBUTES_SYS = 0x00ffff00


class SectionType(IntEnum):
    S_REGULAR = 0x00  # Regular section
    S_ZEROFILL = 0x01  # Zero fill on demand section.
    S_CSTRING_LITERALS = 0x02  # Section with literal C strings
    S_4BYTE_LITERALS = 0x03  # Section with 4 byte literals.
    S_8BYTE_LITERALS = 0x04  # Section with 8 byte literals.
    S_LITERAL_POINTERS = 0x05  # Section with pointers to literals.
    S_NON_LAZY_SYMBOL_POINTERS = 0x06  # Section with non-lazy symbol pointers.
    S_LAZY_SYMBOL_POINTERS = 0x07  # Section with lazy symbol pointers.
    S_SYMBOL_STUBS = 0x08  # Section with symbol stubs, byte size of stub in the Reserved2 field.
    S_MOD_INIT_FUNC_POINTERS = 0x09  # Section with only function pointers for initialization.
    S_MOD_TERM_FUNC_POINTERS = 0x0A  # Section with only function pointers for initialization.
    S_COALESCED = 0x0B  # Section contains symbols that are to be coalesced.
    S_GB_ZEROFILL = 0x0C  # Zero fill on demand section (that can be larger than 4 gigabytes).
    S_INTERPOSING = 0x0D  # Section with only pairs of function pointers for interposing.
    S_16BYTE_LITERALS = 0x0E  # Section with only 16 byte literals.
    S_DTRACE_DOF = 0x0F  # Section contains DTrace Object Format.
    S_LAZY_DYLIB_SYMBOL_POINTERS = 0x10  # Section with lazy symbol pointers to lazy loaded dylibs.
    S_THREAD_LOCAL_REGULAR = 0x11  # Thread local data section.
    S_THREAD_LOCAL_ZEROFILL = 0x12  # Thread local zerofill section.
    S_THREAD_LOCAL_VARIABLES = 0x13  # Section with thread local variable structure data.
    S_THREAD_LOCAL_VARIABLE_POINTERS = 0x14  # Section with pointers to thread local structures.
    S_THREAD_LOCAL_INIT_FUNCTION_POINTERS = 0x15  # Section with thread local variable initialization pointers to functions.


class SectionAttributesUser(IntEnum):
    S_ATTR_PURE_INSTRUCTIONS = 0x80000000  # Section contains only true machine instructions.
    S_ATTR_NO_TOC = 0x40000000  # Section contains coalesced symbols that are not to be in a ranlib table of contents.
    S_ATTR_STRIP_STATIC_SYMS = 0x20000000  # Ok to strip static symbols in this section in files with the MY_DYLDLINK flag.
    S_ATTR_NO_DEAD_STRIP = 0x10000000  # No dead stripping.
    S_ATTR_LIVE_SUPPORT = 0x08000000  # Blocks are live if they reference live blocks.
    S_ATTR_SELF_MODIFYING_CODE = 0x04000000  # Used with i386 code stubs written on by
    S_ATTR_DEBUG = 0x02000000  # A debug section.


class SectionAttributesSys(IntEnum):
    S_ATTR_SOME_INSTRUCTIONS = 0x00000400
    S_ATTR_EXT_RELOC = 0x00000200
    S_ATTR_LOC_RELOC = 0x00000100


CPU_ARCH_MASK = 0xff000000  # Maskfor architecture bits
CPU_ARCH_ABI64 = 0x01000000


class CPUType(IntEnum):
    ANY = -1
    X86 = 7
    X86_64 = X86 | CPU_ARCH_ABI64
    MC98000 = 10
    ARM = 12
    ARM64 = ARM | CPU_ARCH_ABI64
    SPARC = 14
    POWERPC = 18
    POWERPC64 = POWERPC | CPU_ARCH_ABI64


class CPUSubTypeX86(IntEnum):
    ALL = 3
    ARCH1 = 4


class CPUSubTypeX86_64(IntEnum):
    ALL = 3
    H = 8


class CPUSubTypeARM(IntEnum):
    ALL = 0
    V4T = 5
    V6 = 6
    V5 = 7
    V5TEJ = 7
    XSCALE = 8
    V7 = 9
    ARM_V7F = 10
    V7S = 11
    V7K = 12
    V6M = 14
    V7M = 15
    V7EM = 16


class CPUSubTypeARM64(IntEnum):
    ALL = 0
    ARM64E = 2
    ARM64E2 = 0x80000002


class CPUSubTypeSPARC(IntEnum):
    ALL = 0


class CPUSubTypePowerPC(IntEnum):
    ALL = 0
    _601 = 1
    _602 = 2
    _603 = 3
    _603e = 4
    _603ev = 5
    _604 = 6
    _604e = 7
    _620 = 8
    _750 = 9
    _7400 = 10
    _7450 = 11
    _970 = 100


CPU_SUBTYPES = {
    CPUType.X86: CPUSubTypeX86,
    CPUType.X86_64: CPUSubTypeX86_64,
    CPUType.POWERPC: CPUSubTypePowerPC,
    CPUType.ARM: CPUSubTypeARM,
    CPUType.ARM64: CPUSubTypeARM64,
    CPUType.SPARC: CPUSubTypeSPARC
}



#
#  ktool | ktool
#  headers.py
#
#  This file contains the utilities used to create ObjC Header File dumps
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2021.
#

from .dyld import SymbolType
from .objc import ObjCLibrary

from .util import KTOOL_VERSION


class HeaderUtils:

    @staticmethod
    def header_head(library):
        try:
            prefix = "// Headers generated with ktool v" + KTOOL_VERSION + "\n"
            prefix += "// https://github.com/kritantadev/ktool | pip3 install k2l\n"
            prefix += f'// Platform: {library.platform.name} | '
            prefix += f'Minimum OS: {library.minos.x}.{library.minos.y}.{library.minos.z} | '
            prefix += f'SDK: {library.sdk_version.x}.{library.sdk_version.y}.{library.sdk_version.z}\n\n'
            return prefix
        except AttributeError:
            prefix = "// Headers generated with ktool v" + KTOOL_VERSION + "\n"
            prefix += "// https://github.com/kritantadev/ktool | pip3 install k2l\n"
            prefix += "// Issue loading library metadata\n\n"
            return prefix


class TypeResolver:
    def __init__(self, objc_library: ObjCLibrary):
        self.library = objc_library
        classes = []
        self.classmap = {}
        try:
            for sym in objc_library.library.binding_table.symbol_table:
                if sym.type == SymbolType.CLASS:
                    self.classmap[sym.name[1:]] = sym
                    classes.append(sym)
        except AttributeError:
            pass
        self.classes = classes
        self.local_classes = objc_library.classlist
        self.local_protos = objc_library.protolist

    def find_linked(self, classname):
        for local in self.local_classes:
            if local.name == classname:
                return ""
        for local in self.local_protos:
            if local.name == classname[1:-1]:
                return "-Protocol"
        if classname in self.classmap:
            try:
                nam = self.library.library.linked[int(self.classmap[classname].ordinal) - 1].install_name
                if '.dylib' in nam:
                    return None
                return nam
            except Exception as ex:
                pass
        return None


class HeaderGenerator:
    def __init__(self, objc_library):
        self.type_resolver = TypeResolver(objc_library)

        self.library = objc_library
        self.headers = {}

        for objc_class in objc_library.classlist:
            self.headers[objc_class.name + '.h'] = Header(self.type_resolver, objc_class)
        for objc_cat in objc_library.catlist:
            if objc_cat.classname != "":
                self.headers[objc_cat.classname + '+' + objc_cat.name + '.h'] = CategoryHeader(objc_cat)
        for objc_proto in objc_library.protolist:
            self.headers[objc_proto.name + '-Protocol.h'] = ProtocolHeader(objc_proto)

        self.headers[self.library.name + '.h'] = UmbrellaHeader(self.headers)
        self.headers[self.library.name + '-Structs.h'] = StructHeader(objc_library)


class StructHeader:
    def __init__(self, library):
        """
        Scans through structs cached in the ObjCLib's type processor and writes them to a header

        :param library: Library containing structs
        """
        text = ""

        for struct in library.tp.structs.values():
            text += str(struct) + '\n\n'

        self.text = text

    def __str__(self):
        return self.text


class Header:
    def __init__(self, type_resolver, objc_class):
        self.interface = Interface(objc_class)
        self.objc_class = objc_class

        self.type_resolver = type_resolver

        self.forward_declaration_classes = []
        self.forward_declaration_protocols = []

        self.imported_classes = {}
        self.locally_imported_classes = []
        self.locally_imported_protocols = []

        self._get_import_section()

        self.text = self._generate_text()

    def __str__(self):
        return self.text

    def _generate_text(self):
        text = [HeaderUtils.header_head(self.type_resolver.library.library),
                "#ifndef " + self.objc_class.name.upper() + "_H", "#define " + self.objc_class.name.upper() + "_H", ""]

        if len(self.forward_declaration_classes) > 0:
            text.append("@class " + ", ".join(self.forward_declaration_classes) + ";")
        if len(self.forward_declaration_protocols) > 0:
            text.append("@protocol " + ", ".join(self.forward_declaration_protocols) + ";")

        text.append("")

        imported_classes = {}

        for oclass, installname in self.imported_classes.items():
            if '/Frameworks/' in installname:
                nam = installname.split("/")[-1]
                if nam not in imported_classes:
                    imported_classes[nam] = nam
            else:
                imported_classes[oclass] = installname

        for oclass, installname in imported_classes.items():
            text.append(f'#import <{installname.split("/")[-1]}/{oclass}.h>')

        text.append("")

        for oclass in self.locally_imported_classes:
            text.append(f'#import "{oclass}.h"')

        for oprot in self.locally_imported_protocols:
            text.append(f'#import "{oprot}-Protocol.h"')

        text.append("")

        text.append(str(self.interface))
        text.append("")
        text.append("")

        text.append("#endif")

        return "\n".join(text)

    def _get_import_section(self):
        if self.interface.objc_class.superclass != "":
            tp = self.interface.objc_class.superclass.split('_')[-1]
            rt = self.type_resolver.find_linked(tp)
            if rt is None:
                if tp != "id":
                    if tp.startswith('<'):
                        if tp[1:-1] not in self.forward_declaration_protocols:
                            self.forward_declaration_protocols.append(tp[1:-1])
                    elif tp.startswith('NSObject<'):
                        if tp[9:-1] not in self.forward_declaration_protocols:
                            self.forward_declaration_protocols.append(tp[9:-1])
                    else:
                        if tp not in self.forward_declaration_classes:
                            self.forward_declaration_classes.append(tp)
            elif rt == "":
                if tp not in self.locally_imported_classes:
                    self.locally_imported_classes.append(tp)
            else:
                if tp not in self.imported_classes:
                    self.imported_classes[tp] = rt
        for proto in self.interface.objc_class.protocols:
            tname = f'<{proto.name}>'
            rt = self.type_resolver.find_linked(tname)
            if rt == "-Protocol":
                self.locally_imported_protocols.append(proto.name)
            else:
                self.forward_declaration_protocols.append(proto.name)
        for ivar in self.interface.ivars:
            if ivar.is_id:
                tp = ivar.type
                rt = self.type_resolver.find_linked(tp)
                if rt is None:
                    if tp != "id":
                        if tp.startswith('<'):
                            if tp[1:-1] not in self.forward_declaration_protocols:
                                self.forward_declaration_protocols.append(tp[1:-1])
                        elif tp.startswith('NSObject<'):

                            if tp[9:-1] not in self.forward_declaration_protocols:
                                self.forward_declaration_protocols.append(tp[9:-1])
                        else:
                            if tp not in self.forward_declaration_classes:
                                self.forward_declaration_classes.append(tp)
                elif rt == "":
                    if tp not in self.locally_imported_classes:
                        self.locally_imported_classes.append(tp)
                elif rt == "-Protocol":
                    if tp not in self.locally_imported_protocols:
                        self.locally_imported_protocols.append(tp[1:-1])
                else:
                    if tp not in self.imported_classes:
                        self.imported_classes[tp] = rt
        for property in self.interface.properties:
            if property.is_id:
                tp = property.type
                rt = self.type_resolver.find_linked(tp)
                if rt is None:
                    if tp != "id":
                        if tp.startswith('<'):
                            if tp[1:-1] not in self.forward_declaration_protocols:
                                self.forward_declaration_protocols.append(tp[1:-1])
                        elif tp.startswith('NSObject<'):

                            if tp[9:-1] not in self.forward_declaration_protocols:
                                self.forward_declaration_protocols.append(tp[9:-1])
                        else:
                            if tp not in self.forward_declaration_classes:
                                self.forward_declaration_classes.append(tp)
                elif rt == "":
                    if tp not in self.locally_imported_classes:
                        self.locally_imported_classes.append(tp)
                elif rt == "-Protocol":
                    if tp not in self.locally_imported_protocols:
                        self.locally_imported_protocols.append(tp[1:-1])
                else:
                    if tp not in self.imported_classes:
                        self.imported_classes[tp] = rt


class CategoryHeader:
    def __init__(self, objc_category):
        self.category = objc_category

        self.properties = objc_category.properties
        self.methods = objc_category.methods
        self.protocols = objc_category.protocols

        self.interface = CategoryInterface(objc_category)

        self.text = self._generate_text()

    def __str__(self):
        return self.text

    def _generate_text(self):
        text = [HeaderUtils.header_head(self.category.library.library),
                "",
                str(self.interface),
                "",
                ""]

        return "\n".join(text)


class ProtocolHeader:
    def __init__(self, objc_protocol):
        self.protocol = objc_protocol

        self.interface = ProtocolInterface(objc_protocol)

        self.text = self._generate_text()

    def __str__(self):
        return self.text

    def _generate_text(self):
        text = [HeaderUtils.header_head(self.protocol.library.library),
                "",
                str(self.interface),
                "",
                ""]

        return "\n".join(text)


class Interface:
    def __init__(self, objc_class):
        self.objc_class = objc_class

        self.properties = []
        self.methods = []
        self.ivars = []
        self.structs = []

        # just store these so we know not to display them
        self.getters = []
        self.setters = []
        self._process_properties()
        self._process_methods()
        self._process_ivars()

    def __str__(self):
        head = "@interface " + self.objc_class.name + ' : '

        # Decode Superclass Name
        superclass = "NSObject"
        if self.objc_class.superclass != "":  # _OBJC_CLASS_$_UIApplication
            superclass = self.objc_class.superclass.split('_')[-1]

        head += superclass

        # Protocol Implementing Declaration
        if len(self.objc_class.protocols) > 0:
            head += " <"
            for prot in self.objc_class.protocols:
                head += str(prot) + ', '
            head = head[:-2]
            head += '>\n\n'

        # Ivar Declaration
        ivars = ""
        if len(self.ivars) > 0:
            ivars = " {\n"
            for ivar in self.ivars:
                ivars += '    ' + str(ivar) + ';\n'
            ivars += '}\n'

        props = "\n\n"
        for prop in self.properties:
            props += str(prop) + ';'
            if prop.ivarname != "":
                props += ' // ivar: ' + prop.ivarname + '\n'
            else:
                props += '\n'

        meths = "\n\n"
        for i in self.methods:
            if '.cxx_' not in str(i):
                meths += str(i) + ';\n'

        foot = "\n\n@end"
        return head + ivars + props + meths + foot

    def _process_properties(self):
        for property in self.objc_class.properties:
            if not hasattr(property, 'type'):
                continue
            if property.type.lower() == 'bool':
                gettername = 'is' + property.name[0].upper() + property.name[1:]
                self.getters.append(gettername)
            else:
                self.getters.append(property.name)
            if 'readonly' not in property.attributes:
                settername = 'set' + property.name[0].upper() + property.name[1:]
                self.setters.append(settername)
            self.properties.append(property)

    def _process_ivars(self):
        for ivar in self.objc_class.ivars:
            bad = False
            for prop in self.properties:
                if ivar.name == prop.ivarname:
                    bad = True
                    break
            if bad:
                continue
            self.ivars.append(ivar)

    def _process_methods(self):
        for method in self.objc_class.methods:
            bad = False
            for name in self.getters:
                if name in method.sel and ':' not in method.sel:
                    bad = True
                    break
            for name in self.setters:
                if name in method.sel and 'set' in method.sel:
                    bad = True
                    break
            if bad:
                continue
            self.methods.append(method)
        if self.objc_class.metaclass is not None:
            for method in self.objc_class.metaclass.methods:
                self.methods.append(method)


class StructDef:
    def __init__(self, structdef):
        self.structdef = structdef


class CategoryInterface:
    def __init__(self, objc_category):
        self.category = objc_category

        self.properties = self.category.properties
        self.methods = self.category.methods
        self.protocols = self.category.protocols

        self.text = self._generate_text()

    def __str__(self):
        return self.text

    def _generate_text(self):

        head = "@interface "

        head += self.category.classname + " (" + self.category.name + ")"

        # Protocol Implementing Declaration
        if len(self.category.protocols) > 0:
            head += " <"
            for prot in self.category.protocols:
                head += str(prot) + ', '
            head = head[:-2]
            head += '>\n'

        # Ivar Declaration

        props = "\n\n"
        for prop in self.properties:
            props += str(prop) + ';'
            if prop.ivarname != "":
                props += ' // ivar: ' + prop.ivarname + '\n'
            else:
                props += '\n'

        meths = "\n\n"
        for i in self.methods:
            meths += str(i) + ';\n'

        foot = "@end\n"

        return head + props + meths + foot


class ProtocolInterface:
    def __init__(self, protocol):
        self.protocol = protocol

        self.text = self._generate_text()

    def __str__(self):
        return self.text

    def _generate_text(self):
        text = ["@protocol " + self.protocol.name, ""]

        for prop in self.protocol.properties:
            pro = ""
            pro += str(prop) + ';'
            if hasattr(prop, 'ivarname'):
                if prop.ivarname != "":
                    pro += ' // ivar: ' + prop.ivarname + ''
                else:
                    pro += ''
            text.append(pro)

        text.append("")

        for meth in self.protocol.methods:
            text.append(str(meth) + ';')

        text.append("")

        if len(self.protocol.opt_methods) > 0:
            text.append("@optional")
            for meth in self.protocol.opt_methods:
                text.append(str(meth) + ';')

        text.append("@end")

        return "\n".join(text)


class UmbrellaHeader:
    def __init__(self, header_list: dict):
        """
        Generates a header that solely imports other headers

        :param header_list: Dict of headers to be imported
        """
        self.text = "\n\n"
        for header in header_list.keys():
            self.text += "#include \"" + header + "\"\n"

    def __str__(self):
        return self.text



#
#  ktool | ktool
#  generator.py
#
#  Holds some miscellaneous generators for certain filetypes
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2021.
#

import os
from collections import namedtuple

from kmacho.structs import *
from . import log
from .dyld import Dyld, SymbolType
from .macho import Slice
from .objc import ObjCLibrary


class TBDGenerator:
    def __init__(self, library, general=True, objc_lib=None):
        """
        The TBD Generator is a generator that creates TAPI formatted text based stubs for libraries.

        It is currently fairly incomplete, although its output should still be perfectly functional in an SDK.

        After processing, its .dict attribute can be dumped by a TAPI YAML serializer (located in ktool.util) to
            produce a functional .tbd

        :param library: Library being processed
        :type library: Library
        :param general: Should the generator create a .tbd for usage in SDKs?
        :type general: bool
        :param objc_lib: Pass an objc library to the generator. If none is passed it will generate its own
        """
        self.library = library
        self.objc_lib = objc_lib
        self.general = general
        self.dict = self._generate_dict()

    def _generate_dict(self):
        """
        This function simply parses through the library and creates the tbd dict

        :return: The text-based-stub dictionary representation
        """
        tbd = {}
        if self.general:
            tbd['archs'] = ['armv7', 'armv7s', 'arm64', 'arm64e']
            tbd['platform'] = '(null)'
            tbd['install-name'] = self.library.dylib.install_name
            tbd['current-version'] = 1
            tbd['compatibility-version'] = 1

            export_dict = {'archs': ['armv7', 'armv7s', 'arm64', 'arm64e']}

            if len(self.library.allowed_clients) > 0:
                export_dict['allowed-clients'] = self.library.allowed_clients

            symbols = []
            classes = []
            ivars = []

            for sym in self.library.exports.symbols:
                if sym.type == SymbolType.FUNC:
                    symbols.append(sym.name)
                elif sym.type == SymbolType.CLASS:
                    classes.append(sym.name)
                elif sym.type == SymbolType.IVAR:
                    ivars.append(sym.name)

            export_dict['symbols'] = symbols
            export_dict['objc-classes'] = classes
            export_dict['objc-ivars'] = ivars

            tbd['exports'] = [export_dict]
        return tbd


fat_arch_for_slice = namedtuple("fat_arch_for_slice", [
    "slice",
    "cputype",
    "cpusubtype",
    "offset",
    "size",
    "align"
])


class FatMachOGenerator:
    """

    """

    def __init__(self, slices):
        self.slices = slices
        self.fat_archs = []
        pfa = None
        for fat_slice in slices:
            fat_arch_item = self._fat_arch_for_slice(fat_slice, pfa)
            pfa = fat_arch_item
            self.fat_archs.append(fat_arch_item)

        fat_head = bytearray()

        fh = Struct.create_with_values(fat_header, [b'\xCA\xFE\xBA\xBE', len(self.fat_archs)], "big")

        fat_head += fh.raw

        for fat_arch_item in self.fat_archs:
            fa = Struct.create_with_values(fat_arch,
                                           [fat_arch_item.cputype, fat_arch_item.cpusubtype, fat_arch_item.offset,
                                            fat_arch_item.size, fat_arch_item.align], "big")
            fat_head += fa.raw

        self.fat_head = fat_head

    @staticmethod
    def _fat_arch_for_slice(fat_slice: Slice, previous_fat_arch):
        """
        :param fat_slice: Fat slice
        :type fat_slice: Slice
        :param previous_fat_arch: Previous item returned by this func, or None if first.
        :type previous_fat_arch: fat_arch_for_slice
        :return: fat_arch_for_slice item.
        :rtype: fat_arch_for_slice
        """
        lib = Dyld.load(fat_slice)
        cputype = lib.macho_header.dyld_header.cputype
        cpu_subtype = lib.macho_header.dyld_header.cpu_subtype

        if len(fat_slice.macho_file.slices) > 1:
            size = fat_slice.arch_struct.size
            align = pow(2, fat_slice.arch_struct.align)
            align_directive = fat_slice.arch_struct.align
        else:
            f = fat_slice.macho_file.file_object
            old_file_position = f.tell()
            f.seek(0, os.SEEK_END)
            size = f.tell()
            f.seek(old_file_position, os.SEEK_SET)

            if cputype == 16777228:
                align = pow(2, 0xe)
                align_directive = 0xe
            elif cputype == 16777223:
                align = pow(2, 0xc)
                align_directive = 0xc
            else:
                # TODO: implement other alignment directives and stuff
                print(cputype)
                raise AssertionError("not yet implemented")
        if previous_fat_arch is None:
            offset = align
        else:
            offset = 0
            while True:
                offset += align
                if offset > previous_fat_arch.offset + previous_fat_arch.size:
                    break

        log.debug(f'Create arch with offset {hex(offset)} and size {hex(size)}')

        return fat_arch_for_slice(fat_slice, cputype, cpu_subtype, offset, size, align_directive)



from setuptools import setup

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(name='k2l',
      version='0.15.5',
      description='Static mach-o/img4 analysis tool.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      python_requires='>=3.6',
      author='kritanta',
      url='https://github.com/kritantadev/ktool',
      install_requires=['pyaes', 'kimg4', 'Pygments', 'packaging'],
      packages=['kmacho', 'ktool'],
      package_dir={
            'kmacho': 'src/kmacho',
            'ktool': 'src/ktool'
      },
      classifiers=[
            'Programming Language :: Python :: 3',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent'
      ],
      scripts=['bin/ktool']
      )
